{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "[Day 21](https://adventofcode.com/2019/day/21) puzzle input is an IntCode program (mine is [here](puzzle_input/day21.txt)) which represents a networked computer as part of a bigger network. Part 1 involves connecting a number of these network computers and finding the first "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [],
   "source": [
    "opcode_add = 1\n",
    "opcode_mul = 2\n",
    "opcode_read = 3\n",
    "opcode_write = 4\n",
    "opcode_jump_true = 5\n",
    "opcode_jump_false = 6\n",
    "opcode_lt = 7\n",
    "opcode_eq = 8\n",
    "opcode_rebase = 9\n",
    "\n",
    "opcode_terminate = 99\n",
    "\n",
    "mode_position = 0\n",
    "mode_immediate = 1\n",
    "mode_relative = 2\n",
    "\n",
    "class IntCodeCpu:\n",
    "    def __init__(self, memory_image):\n",
    "        self.memory = [ x for x in memory_image ] # copy memory image, in case it's reused\n",
    "        self.stalled = True\n",
    "        self.input_buffer = None\n",
    "        self.output_buffer = None\n",
    "        self.pc = 0\n",
    "        self.initialise_opcodes()\n",
    "        self.offset = 0\n",
    "        self.done = False\n",
    "        \n",
    "    def start(self, input_buffer, output_buffer, noun=None, verb=None):\n",
    "        self.input_buffer = input_buffer\n",
    "        self.output_buffer = output_buffer\n",
    "\n",
    "        if noun:\n",
    "            self.memory[1] = noun\n",
    "        if verb:\n",
    "            self.memory[2] = verb\n",
    "           \n",
    "        return self.run()\n",
    "            \n",
    "            \n",
    "    def run(self):\n",
    "        instr = self.memory[self.pc]\n",
    "        self.stalled = False\n",
    "\n",
    "        while int(instr) != opcode_terminate and not self.stalled:\n",
    "            (op, modes) = self.decode_instr(instr)\n",
    "            self.pc = op(modes)\n",
    "            instr = self.memory[self.pc]\n",
    "\n",
    "        return self.memory[0]\n",
    "\n",
    "    #-HELPERS-----------------------------\n",
    "    def try_pop_mode(self, modes):\n",
    "        if len(modes) == 0:\n",
    "            return 0\n",
    "    \n",
    "        return modes.pop()\n",
    "    \n",
    "    def resize_memory(self, target_addr):\n",
    "        self.memory += ([0] * (1 + target_addr - len(self.memory)))\n",
    "    \n",
    "    #-DECODE-INSTRUCTIONS-----------------\n",
    "    def initialise_opcodes(self):\n",
    "        self.opcodes = {\n",
    "            opcode_add: self.op_add,\n",
    "            opcode_mul: self.op_mul,\n",
    "            opcode_read: self.op_read,\n",
    "            opcode_write: self.op_write,\n",
    "            opcode_jump_true: self.op_jump_true,\n",
    "            opcode_jump_false: self.op_jump_false,\n",
    "            opcode_lt: self.op_lt,\n",
    "            opcode_eq: self.op_eq,\n",
    "            opcode_rebase: self.op_rebase\n",
    "        }    \n",
    "    \n",
    "    def decode_instr(self, instr):\n",
    "        instr = str(instr)\n",
    "        opcode = self.decode_op(instr)\n",
    "        modes = self.decode_modes(instr)\n",
    "\n",
    "        if not (opcode in self.opcodes):\n",
    "            raise Exception(f\"Invalid opcode {opcode}\")\n",
    "\n",
    "        return (self.opcodes[opcode], modes)    \n",
    "    \n",
    "    def decode_op(self, instr):\n",
    "        if len(instr) > 2:\n",
    "            return int(instr[-2:])\n",
    "        return int(instr)\n",
    "    \n",
    "    def decode_modes(self, instr):\n",
    "        if len(instr) > 2:\n",
    "            return [ int(d) for d in instr[:-2]]\n",
    "        return []\n",
    "    \n",
    "    #-MICRO-OPS---------------------------\n",
    "    def uop_read(self, value, mode):\n",
    "        if mode == mode_position:\n",
    "            if value >= len(self.memory):\n",
    "                self.resize_memory(value)        \n",
    "\n",
    "            return int(self.memory[value])\n",
    "\n",
    "        elif mode == mode_relative:\n",
    "            if self.offset + value >= len(self.memory):\n",
    "                self.resize_memory(self.offset + value)     \n",
    "\n",
    "            return int(self.memory[self.offset + value])\n",
    "\n",
    "        elif mode == mode_immediate:\n",
    "            return int(value)\n",
    "\n",
    "        else:\n",
    "            raise Exception(\"UNKNOWN MODE\")\n",
    "\n",
    "    def uop_write(self, dst, value, mode):\n",
    "        if mode == mode_position:\n",
    "            if dst >= len(self.memory):\n",
    "                self.resize_memory(dst)\n",
    "            self.memory[dst] = value\n",
    "\n",
    "        elif mode == mode_relative:\n",
    "            if self.offset + dst >= len(self.memory):\n",
    "                self.resize_memory(self.offset + dst)     \n",
    "\n",
    "            self.memory[self.offset + dst] = value\n",
    "\n",
    "        elif mode == mode_immediate:\n",
    "            raise Exception(f\"cannot write {value} to literal {dst}\")\n",
    "\n",
    "\n",
    "    def uop_cond_jump(self, modes, cond):\n",
    "        param_mode = self.try_pop_mode(modes)\n",
    "        param_raw = int(self.memory[self.pc + 1])\n",
    "        param = self.uop_read(param_raw, param_mode)\n",
    "\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "        dest_raw = int(self.memory[self.pc + 2])\n",
    "        dest = self.uop_read(dest_raw, dest_mode)\n",
    "\n",
    "        if cond(param):\n",
    "            return dest\n",
    "\n",
    "        return self.pc + 3\n",
    "\n",
    "    def uop_cmp(self, modes, cmp):\n",
    "\n",
    "        param0_mode = self.try_pop_mode(modes)\n",
    "        param0_raw = int(self.memory[self.pc + 1])\n",
    "        param0 = self.uop_read(param0_raw, param0_mode)\n",
    "\n",
    "        param1_mode = self.try_pop_mode(modes)\n",
    "        param1_raw = int(self.memory[self.pc + 2])\n",
    "        param1 = self.uop_read(param1_raw, param1_mode)\n",
    "\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "        dest = int(self.memory[self.pc + 3])\n",
    "\n",
    "        if cmp(param0, param1):\n",
    "            self.uop_write(dest, 1, dest_mode)\n",
    "        else:\n",
    "            self.uop_write(dest, 0, dest_mode)\n",
    "\n",
    "        return self.pc + 4\n",
    "        \n",
    "    \n",
    "    #-OPCODES-----------------------------\n",
    "    def op_add(self, modes):\n",
    "        arg0_mode = self.try_pop_mode(modes)\n",
    "        arg1_mode = self.try_pop_mode(modes)\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "\n",
    "        arg0_raw = int(self.memory[self.pc + 1])\n",
    "        arg1_raw = int(self.memory[self.pc + 2])\n",
    "        dest = int(self.memory[self.pc + 3])\n",
    "\n",
    "        arg0 = self.uop_read(arg0_raw, arg0_mode)\n",
    "        arg1 = self.uop_read(arg1_raw, arg1_mode)\n",
    "\n",
    "        self.uop_write(dest, str(int(arg0) + int(arg1)), dest_mode)\n",
    "        return self.pc + 4\n",
    "\n",
    "    def op_mul(self, modes):\n",
    "        arg0_mode = self.try_pop_mode(modes)\n",
    "        arg1_mode = self.try_pop_mode(modes)\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "\n",
    "        arg0_raw = int(self.memory[self.pc + 1])\n",
    "        arg1_raw = int(self.memory[self.pc + 2])\n",
    "        dest = int(self.memory[self.pc + 3])\n",
    "\n",
    "        arg0 = self.uop_read(arg0_raw, arg0_mode)\n",
    "        arg1 = self.uop_read(arg1_raw, arg1_mode)\n",
    "\n",
    "        self.uop_write(dest, str(int(arg0) * int(arg1)), dest_mode)    \n",
    "        return self.pc + 4\n",
    "\n",
    "\n",
    "    def op_read(self, modes):\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "        dest = int(self.memory[self.pc + 1])\n",
    "        \n",
    "        # if the input buffer is empty, we should \"stall\" and \n",
    "        # resume later\n",
    "        if not self.input_buffer:\n",
    "            self.stalled = True\n",
    "            return self.pc\n",
    "        \n",
    "        val = self.input_buffer.pop()\n",
    "\n",
    "        self.uop_write(dest, str(val), dest_mode)\n",
    "\n",
    "        return self.pc + 2\n",
    "\n",
    "    def op_write(self, modes):\n",
    "        src_mode = self.try_pop_mode(modes)\n",
    "        src_raw = int(self.memory[self.pc + 1])\n",
    "        src = self.uop_read(src_raw, src_mode)\n",
    "\n",
    "        self.output_buffer.append(src)\n",
    "\n",
    "        return self.pc + 2\n",
    "\n",
    "    def op_jump_true(self, modes):\n",
    "        return self.uop_cond_jump(modes, lambda x: x != 0)\n",
    "\n",
    "\n",
    "    def op_jump_false(self, modes):\n",
    "        return self.uop_cond_jump(modes, lambda x: x == 0)\n",
    "\n",
    "    def op_lt(self, modes):\n",
    "        return self.uop_cmp(modes, lambda x, y: x < y)\n",
    "\n",
    "    def op_eq(self, modes):\n",
    "        return self.uop_cmp(modes, lambda x, y: x == y)\n",
    "\n",
    "    def op_rebase(self, modes):    \n",
    "        param_mode = self.try_pop_mode(modes)\n",
    "        param_raw = int(self.memory[self.pc + 1])\n",
    "        param = self.uop_read(param_raw, param_mode)\n",
    "\n",
    "        self.offset += param\n",
    "\n",
    "        return self.pc + 2"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [],
   "source": [
    "raw_code = open(\"puzzle_input/day21.txt\", \"r\").read()\n",
    "code = raw_code.split(\",\")\n",
    "\n",
    "def preprocess_ascii_input(instructions):\n",
    "    int_instructions = [ ord(c) for c in instructions ]\n",
    "    return int_instructions + [ ord(\"\\n\") ]\n",
    "\n",
    "def skip_line(line):\n",
    "    return len(line) == 0 or line[0] == \"#\"\n",
    "\n",
    "def run_script(script):\n",
    "    output_buffer = []\n",
    "    cpu = IntCodeCpu(code)\n",
    "    script_lines = [ preprocess_ascii_input(script_line) for script_line in script.split(\"\\n\") if not skip_line(script_line.strip()) ]\n",
    "    \n",
    "    from pprint import pprint\n",
    "    output = []\n",
    "    for i, script_line in enumerate(script_lines):\n",
    "        print(i, script_line)\n",
    "        cpu.start(script_line[::-1], output_buffer)\n",
    "        \n",
    "    maybe_return_value = output_buffer[-1]\n",
    "    if (maybe_return_value >= 0 and maybe_return_value < 256):\n",
    "        # the output is printable ascii meaning that we\n",
    "        # have hit a snag - we should print the output \n",
    "        # to see the last dying moments of our dear\n",
    "        # springbot\n",
    "        output = \"\".join([ chr(c) for c in output_buffer ])\n",
    "        print(output)\n",
    "    else:\n",
    "        print(\"retval = \", maybe_return_value)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "0 [78, 79, 84, 32, 67, 32, 84, 10]\n",
      "1 [65, 78, 68, 32, 68, 32, 84, 10]\n",
      "2 [79, 82, 32, 84, 32, 74, 10]\n",
      "3 [78, 79, 84, 32, 65, 32, 84, 10]\n",
      "4 [79, 82, 32, 84, 32, 74, 10]\n",
      "5 [87, 65, 76, 75, 10]\n",
      "retval =  19354392\n"
     ]
    }
   ],
   "source": [
    "# part 1\n",
    "# reached via some very simple heuristics - if there is\n",
    "# an gap upcoming just before jump target (D) we should jump:\n",
    "#\n",
    "#    @\n",
    "#    #??.# \n",
    "#     ABCD\n",
    "#\n",
    "# and if there's a gap immediately in front of us, we should jump\n",
    "#\n",
    "#    @\n",
    "#    #.???\n",
    "#     ABCD\n",
    "# \n",
    "# otherwise, we'll walk\n",
    "\n",
    "script_part_1 = \"\"\"\n",
    "# if there's a gap *just* before jump target, jump\n",
    "NOT C T\n",
    "AND D T\n",
    "OR T J\n",
    "\n",
    "# if there's a gap right in front of us, jump\n",
    "NOT A T\n",
    "OR T J\n",
    "\n",
    "# fingers crossed\n",
    "WALK\"\"\"\n",
    "\n",
    "run_script(script_part_1)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "0 [78, 79, 84, 32, 67, 32, 74, 10]\n",
      "1 [65, 78, 68, 32, 68, 32, 74, 32, 10]\n",
      "2 [65, 78, 68, 32, 72, 32, 74, 10]\n",
      "3 [78, 79, 84, 32, 66, 32, 84, 32, 10]\n",
      "4 [65, 78, 68, 32, 68, 32, 84, 32, 10]\n",
      "5 [79, 82, 32, 84, 32, 74, 10]\n",
      "6 [78, 79, 84, 32, 65, 32, 84, 32, 10]\n",
      "7 [79, 82, 32, 84, 32, 74, 10]\n",
      "8 [82, 85, 78, 10]\n",
      "retval =  1139528802\n"
     ]
    }
   ],
   "source": [
    "# part 2\n",
    "# \n",
    "# more tricky, this time we need to be more careful about our\n",
    "# fallback of \"otherwise, we'll walk\"\n",
    "#\n",
    "#    ......@..........\n",
    "#    #####.##.##..####\n",
    "#           ABCDEFGH\n",
    "#\n",
    "# here the existing program believes we can go ahead - but we\n",
    "# end up being forced to jump. we need to jump before this but\n",
    "# how do we decide this? a naive guess would be \"if there's yet\n",
    "# another gap upcoming whose last tile is F then jump\"\n",
    "\n",
    "script_part_2 = \"\"\"\n",
    "# if there's a gap just before jump target AND\n",
    "# we can continue jumping if we need, then jump\n",
    "NOT C J\n",
    "AND D J \n",
    "AND H J\n",
    "\n",
    "# however make sure NOT to walk a step when \n",
    "# being forced to jump would land us in gap\n",
    "NOT B T \n",
    "AND D T \n",
    "OR T J\n",
    "\n",
    "# if there's a gap right in front of us, jump\n",
    "NOT A T \n",
    "OR T J\n",
    "\n",
    "# fingers crossed\n",
    "RUN\n",
    "\"\"\"        \n",
    "\n",
    "run_script(script_part_2)"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.9.0"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
